'use strict';

const Azure = require('azure-storage');
const Generator = Azure.TableUtilities.entityGenerator;
const MemoryStream = require('memorystream');
const { Readable } = require('stream');

/**
 * Handles CRUD operations with Azure Blob storage.
 *
 * EXAMPLE:
 * (to run it, the container must already be created - you can do it via
 * Azure Portal)

  (async () => {
    // Connect to service:
    const Blobs = require('./azure-blob-storage');
    global.blobs = new Blobs(
      process.env.AZURE_STORAGE_ACCOUNT, process.env.AZURE_STORAGE_ACCESS_KEY);

    // Assume that the container 'my-container' is already created.
    let container = 'my-container';
    let testContent = 'Test! ' + Date.now();

    // Write file:
    console.log('Testing file write...');
    await global.blobs.writeFile(container, 'test.txt', testContent);

    // Read file:
    console.log('Testing file read...');
    let readData = await global.blobs.readFile(container, 'test.txt');

    console.log('Read data: ', readData.toString('utf8'));
  })();

 */
class AzureBlobStorage {
  constructor(storageAccount, storageAccessKey) {
    try {
      this.service = Azure.createBlobService(storageAccount, storageAccessKey);
      this.storageAccount = storageAccount;
    } catch (error) {
      console.log('Error initializing Azure Blob Storage: ', error);
      throw error;
    }
  }

  /**
   * Checks whether all specified containers exist, and if not, creates them.
   * @param {object} containers - Object containing container names and types.
   */
  async initializeContainers(containers) {
    for (let container of Object.values(containers)) {
      console.log(`Initializing container ${container.name} in Azure storage...`);

      let options = {};
      if (container.type === 'public') {
        options.publicAccessLevel = 'blob'; // has a public URL
      } else {
        // do nothing, default is private
      }

      await new Promise((resolve, reject) => {
        this.service.createContainerIfNotExists(container.name, options, (error, result) => {
          if (error) {
            reject(error);
          } else {
            resolve(result);
          }
        });
      });
    }
  }

  /**
   * Writes a file to blob.
   * @param {string} container - The container name.
   * @param {string} filename
   * @param {string|Buffer} content - The contents of the file to write
   * @returns {Promise<object>}
   */
  writeFile(container, filename, content) {
    return new Promise((resolve, reject) => {
      let [dataStream, length] = stringToStream(content);

      this.service.createBlockBlobFromStream(container, filename, dataStream, length, (error, result) => {
        if (error) {
          reject(error);
        } else {
          resolve(result);
        }
      });
    });
  }

  /**
   * Append a string or buffer to the existing blob. 
   * @param {string} container - The Azure blob service container name.
   * @param {string} filename - The name of the blob.
   * @param {string|Buffer} content - The contents to append (string or Buffer).
   */
  append(container, filename, content) {
    return new Promise((resolve, reject) => {
      let [dataStream, length] = stringToStream(content);

      this.service.appendBlockFromStream(container, filename, dataStream, length, (error, result) => {
        if (error) {
          reject(error);
        } else {
          resolve(result);
        }
      });
    });
  }

  /**
   * Reads a file from Azure Blob storage and returns it as a buffer.
   * @param {string} container
   * @param {string} filename
   * @returns {Promise<Buffer>} The file data.
   */
  readFile(container, filename) {
    return new Promise((resolve, reject) => {
      let writableStream = new MemoryStream();
      try {
        let downloadedData = new Promise((resolve2, reject2) => {
          this.service.getBlobToStream(container, filename, writableStream, async (error, result, response) => {
            if (error) {
              reject2(error);
            } else {
              let data = await readStreamAsString(writableStream);
              resolve2(data);
            }
          });
        });

        resolve(downloadedData);
      } catch (outerError) {
        reject(outerError);
      }
    });
  }

  /**
   * Checks that the specified file exists.
   * @param {string} container
   * @param {string} filename
   */
  exists(container, filename) {
    return new Promise((resolve, reject) => {
      try {
        this.service.doesBlobExist(container, filename, (error, result) => {
          if (error) {
            reject(error);
          } else {
            resolve(result.exists);
          }
        });
      } catch (error) {
        reject(error);
      }
    });
  }

  /**
   * Returns a public URL for downloading a file. Will return a link even
   * if the container is non-public, but the link won't work.
   * @param {string} container
   * @param {anystring} filename
   */
  getPublicDownloadLink(container, filename) {
    return `https://${this.storageAccount}.blob.core.windows.net/${container}/${filename}`;
  }
}


/**
 * Handles CRUD operations for Azure Tables, a cheaper and faster alternative
 * to SQL databases when relational databases are not needed.
 */
class AzureTableStorage {
  constructor(storageAccount, storageAccessKey) {
    try {
      this.service = Azure.createTableService(storageAccount, storageAccessKey);
      this.storageAccount = storageAccount;
    } catch (error) {
      console.log('Error initializing Azure Table Storage: ', error);
      throw error;
    }
  }

  /**
   * Creates tables if they don't exist yet.
   * @param {string[]} tables - Table names.
   */
  async initializeTables(tables) {
    for (let table of tables) {
      await new Promise((resolve, reject) => {
        console.log(`Initializing table ${table}...`);
        this.service.createTableIfNotExists(table, (error, result, response) => {
          if (error) {
            reject(error);
          } else {
            resolve(error);
          }
        });
      });
    }
  }

  /**
   * Reads a specified row from the table.
   * @param {any} table
   * @param {any} partition
   * @param {any} key
   */
  async retrieveEntity(table, partition, key) {
    return await new Promise((resolve, reject) => {
      // Numbers are not allowed:
      partition = partition.toString();
      key = key.toString();

      this.service.retrieveEntity(table, partition, key, (error, result, response) => {
        if (error) {
          reject(error);
        } else {
          // Convert the response format where values are stored in the _
          // key into a normal object format:
          resolve(azureTablesRowToObject(result));
        }
      });
    });
  }

  /**
   * Retrieves all entities from a partition.
   * @param {string} table
   * @param {string} partition
   */
  async retrieveAllEntitiesByPartitionKey(table, partition) {
    return await new Promise((resolve, reject) => {
      let query = new Azure.TableQuery().where('PartitionKey eq ?', partition);

      this.service.queryEntities(table, query, null, async (error, result, response) => {
        if (error) {
          reject(error);
        } else {
          if (result.continuationToken) {
            throw new Error('TODO FIXME - continuation tokens not implemented yet.');
          } else {
            let data = await azureTablesArrayToObject(result.entries);
            resolve(data);
          }
        }
      });
    });
  }

  /**
   * Adds or replaces an entity where all values are strings.
   * @param {string} table - Table name.
   * @param {string} partition - Partition key.
   * @param {string} key - Row key (row ID).
   * @param {object} data - The data object to be saved, where object keys are
   * column names.
   */
  async saveStringOnlyEntity(table, partition, key, data) {
    let entity = {
      PartitionKey: Generator.String(partition.toString()),
      RowKey: Generator.String(key.toString()),
    };

    for (let dataKey of Object.keys(data)) {
      entity[dataKey] = Generator.String(data[dataKey].toString ? data[dataKey].toString() : '');
    }

    return await new Promise((resolve, reject) => {
      this.service.insertOrReplaceEntity(table, entity, (error, result, response) => {
        if (error) {
          reject(error);
        } else {
          resolve(error);
        }
      });
    });
  }

  /**
   * Deletes the row with the specified key.
   * @param {string} table
   * @param {string} partition
   * @param {string} key - The row key of the entity to be deleted.
   */
  async deleteEntity(table, partition, key) {
    return await new Promise((resolve, reject) => {
      this.service.deleteEntity(
        table,
        {
          PartitionKey: partition,
          RowKey: key,
        },
        (error, success, response) => {
          if (!error && success) {
            resolve(success);
          } else {
            reject(response);
          }
        }
      );
    });
  }

  /**
   * Queries a table based on the column values.
   * @param {string} table
   * @param {string} partition
   * @param {string[][]} conditions - An array of arrays, where each condition
   * is an array containing two or more elements, for example ['userName eq ?', 'Anton']
   * or ['hitCount >= ? and hitCount <= ?', '2', '100'].
   */
  async query(table, partition, conditions) {
    return await new Promise((resolve, reject) => {
      let query = new Azure.TableQuery()
        .where('PartitionKey eq ?', partition);

      for (let condition of conditions) {
        query = query.and(...condition);
      }

      this.service.queryEntities(table, query, null, async (error, result, response) => {
        if (error) {
          reject(error);
        } else {
          let processedData = await azureTablesArrayToObject(result.entries);
          resolve(processedData);
        }
      }); 
    });
  }
}

/**
 * Converts the array of objects, each of which is in Azure Table's internal
 * format, into an array of normal JavaScript objects where each key is a
 * column name.
 * @param {object[]} data
 */
async function azureTablesArrayToObject(data) {
  let result = [];
  for (let row of data) {
    result.push(azureTablesRowToObject(row));
  }
  return result;
}

function azureTablesRowToObject(row) {
  let data = {};
  Object.keys(row).forEach(column => {
    if (column !== '.metadata') {
      data[column] = row[column]['_'];
    }
  });
  return data;
}

/**
 * Converts a stream into a string.
 * @param {Writable} stream - Any stream, e.g. a MemoryStream.
 */
function readStreamAsString(stream) {
  let bufferParts = [];

  // TODO FIXME - it may happen that the promise never resolves?
  // Should we have a timeout?
  return new Promise((resolve, reject) => {
    stream.on('data', (buffer) => {
      bufferParts.push(buffer);
    });

    stream.on('end', () => {
      resolve(Buffer.concat(bufferParts));
    });

    stream.on('error', (error) => {
      reject(error);
    });
  });
}

/**
 * Converts a buffer or string into a readable stream.
 * @param {string|Buffer} contentStringOrBuffer
 */
function stringToStream(contentStringOrBuffer) {
  let dataStream = new Readable();

  // For some reason, if content is a string, then content.length
  // will be incorrect and Azure will create an incomplete file. So convert
  // the string into the buffer.
  let contentBuffer = Buffer.from(contentStringOrBuffer);

  dataStream.push(contentBuffer);
  dataStream.push(null); // end the stream

  return [dataStream, contentBuffer.length];
}

module.exports = {
  Blobs: AzureBlobStorage,
  Tables: AzureTableStorage,
};